/***************************************************************************
 *                                                                         *
 *   kemboss - Graphical user interface for EMBOSS                         *
 *                                                                         *
 *   Copyright (C) 2007-2013 Tobias Schwarz (mail@tobias-schwarz.net)      *
 *                                                                         *
 *   Based on the EMBOSS.kaptn scripts by Thomas Siegmund                  *
 *                                                                         *
 *   This program is free software; you can redistribute and/or modify     *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version, and provided that the above       *
 *   copyright and permission notice is included with all distributed      *
 *   copies of this or derived software.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 **************************************************************************/

#include <QStyleFactory>
#include <QMessageBox>
#include <QSettings>
#include <QFileDialog>
#include <QSslSocket>

#include "configDlg.h"
#include "kembossWindowClass.h"

configDlg::configDlg(kemboss *parent, kembossSettings *fksettings) : QDialog(parent)
{
   setupUi(this);
   StyleComboBox->addItem("default");
   StyleComboBox->addItems(QStyleFactory::keys());
   mainKembossWindow = parent;
   ksettings = fksettings;

   readSettings();
   // Reads the settings of the configuration dialog (Position on screen, size, etc.). The kemboss settings are stored in the ksettings object

   KembossDataLineEdit->setText(ksettings->kembossData());
   KembossDefaultSettingsLineEdit->setText(ksettings->kembossDefaultSettings());
   EmbossBinLineEdit->setText(ksettings->embossBin());
   EmbossHelpLineEdit->setText(ksettings->embossHelp());
   OnlineDocCheckBox->setChecked(ksettings->useOnlineDoc());
   OnlineDocSourceComboBox->setCurrentText(ksettings->onlineDocSource());
   if (ksettings->useOnlineDoc() == false)
       OnlineDocSourceComboBox->setEnabled(false);
   EmbossDataLineEdit->setText(ksettings->embossData());
   OpenLastSettingsCheckBox->setChecked(ksettings->openLastSettings());
   CloseProgressWindowCheckBox->setChecked(ksettings->autoCloseProgressWindow());
   ShowOutputTabCheckBox->setChecked(ksettings->autoShowOutputTab());
   AutoDeleteExistingOutputFileCheckBox->setChecked(ksettings->autoDeleteExistingOutputFile());
   FontComboBox->setCurrentFont(ksettings->textEditFont());
   FontSizeSpinBox->setValue(ksettings->textEditFont().pointSize());
   WrapLinesCheckBox->setChecked(ksettings->wrapLines());

   // Set kemboss client options
   LocalEmbossInstallation->setChecked(!ksettings->remoteEmboss());
   RemoteEmbossInstallation->setChecked(ksettings->remoteEmboss());
   HostnameLineEdit->setText(ksettings->hostname());
   PortSpinBox->setValue(ksettings->port());
   UserNameLineEdit->setText(ksettings->userName());
   if (!ksettings->userPwHash().isEmpty()) PwLineEdit->setText("xxxxxxxx");
   EmailLineEdit->setText(ksettings->userEmail());

   // Set color coding colors
   CysteinPushButton->setPalette(QPalette(ksettings->cystein.foreground().color()));
   ApolarAAPushButton->setPalette(QPalette(ksettings->apolar_aa.foreground().color()));
   PolarAAPushButton->setPalette(QPalette(ksettings->polar_aa.foreground().color()));
   AcidicAAPushButton->setPalette(QPalette(ksettings->acidic_aa.foreground().color()));
   BasicAAPushButton->setPalette(QPalette(ksettings->basic_aa.foreground().color()));
   DibasicAAPushButton->setPalette(QPalette(ksettings->dibasic_aa.foreground().color()));
   ANucPushButton->setPalette(QPalette(ksettings->a_nuc.foreground().color()));
   TNucPushButton->setPalette(QPalette(ksettings->t_nuc.foreground().color()));
   GNucPushButton->setPalette(QPalette(ksettings->g_nuc.foreground().color()));
   CNucPushButton->setPalette(QPalette(ksettings->c_nuc.foreground().color()));
   CysteinPushButton->setAutoFillBackground(true);
   ApolarAAPushButton->setAutoFillBackground(true);
   PolarAAPushButton->setAutoFillBackground(true);
   AcidicAAPushButton->setAutoFillBackground(true);
   BasicAAPushButton->setAutoFillBackground(true);
   DibasicAAPushButton->setAutoFillBackground(true);
   ANucPushButton->setAutoFillBackground(true);
   TNucPushButton->setAutoFillBackground(true);
   GNucPushButton->setAutoFillBackground(true);
   CNucPushButton->setAutoFillBackground(true);

   fontHasChanged = false;
   connect(FontComboBox, SIGNAL(currentIndexChanged(int)), this, SLOT(fontChanged()));
   connect(FontSizeSpinBox, SIGNAL(valueChanged(int)), this, SLOT(fontChanged()));
   connect(WrapLinesCheckBox, SIGNAL(stateChanged(int)), this, SLOT(fontChanged()));

   embassyHasChanged = false;
   connect(PhylipnewCheckBox, SIGNAL(stateChanged(int)), this, SLOT(embassyChanged()));

   remoteServerHasChanged = false;
   connect(RemoteEmbossInstallation, SIGNAL(clicked()), this, SLOT(remoteServerChanged()));
   connect(HostnameLineEdit, SIGNAL(textEdited(QString)), this, SLOT(remoteServerChanged()));
   connect(PortSpinBox, SIGNAL(valueChanged(int)), this, SLOT(remoteServerChanged()));
   connect(UserNameLineEdit, SIGNAL(textEdited(QString)), this, SLOT(remoteServerChanged()));
   connect(PwLineEdit, SIGNAL(textEdited(QString)), this, SLOT(remoteServerChanged()));
   connect(EmailLineEdit, SIGNAL(textEdited(QString)), this, SLOT(remoteServerChanged()));
   connect(ButtonBox, SIGNAL(helpRequested()), this, SLOT(helpRequested()));
}

void configDlg::accept()
{
   ksettings->setkembossData(KembossDataLineEdit->text());
   ksettings->setkembossDefaultSettings(KembossDefaultSettingsLineEdit->text());
   ksettings->setembossBin(EmbossBinLineEdit->text());
   ksettings->setembossHelp(EmbossHelpLineEdit->text());
   ksettings->setuseOnlineDoc(OnlineDocCheckBox->checkState());
   ksettings->setonlineDocSource(OnlineDocSourceComboBox->currentText());
   ksettings->setembossData(EmbossDataLineEdit->text());
   ksettings->setopenLastSettings(OpenLastSettingsCheckBox->checkState());
   ksettings->setautoCloseProgressWindow(CloseProgressWindowCheckBox->checkState());
   ksettings->setautoShowOutputTab(ShowOutputTabCheckBox->checkState());
   ksettings->setautoDeleteExistingOutputFile(AutoDeleteExistingOutputFileCheckBox->checkState());
   ksettings->settextEditFont(FontComboBox->currentFont().family(), FontSizeSpinBox->value());
   ksettings->setwrapLines(WrapLinesCheckBox->checkState());

   ksettings->setRemoteEmboss(RemoteEmbossInstallation->isChecked());
   ksettings->setHostname(HostnameLineEdit->text());
   ksettings->setPort(PortSpinBox->value());
   ksettings->setUserName(UserNameLineEdit->text());
   if (!PwLineEdit->text().isEmpty() && PwLineEdit->text() != "xxxxxxxx") ksettings->setUserPassword(PwLineEdit->text());
   ksettings->setUserEmail(EmailLineEdit->text());

   ksettings->cystein.setForeground(CysteinPushButton->palette().color(QPalette::Button));
   ksettings->apolar_aa.setForeground(ApolarAAPushButton->palette().color(QPalette::Button));
   ksettings->polar_aa.setForeground(PolarAAPushButton->palette().color(QPalette::Button));
   ksettings->acidic_aa.setForeground(AcidicAAPushButton->palette().color(QPalette::Button));
   ksettings->basic_aa.setForeground(BasicAAPushButton->palette().color(QPalette::Button));
   ksettings->dibasic_aa.setForeground(DibasicAAPushButton->palette().color(QPalette::Button));
   ksettings->a_nuc.setForeground(ANucPushButton->palette().color(QPalette::Button));
   ksettings->t_nuc.setForeground(TNucPushButton->palette().color(QPalette::Button));
   ksettings->g_nuc.setForeground(GNucPushButton->palette().color(QPalette::Button));
   ksettings->c_nuc.setForeground(CNucPushButton->palette().color(QPalette::Button));

   ksettings->checkConfiguration();
   ksettings->checkKembossDirectories();   
   if (ksettings->remoteEmboss()) {
       if (remoteServerHasChanged) {
           ksettings->checkServer(); // Checks the settings for consistency
           mainKembossWindow->checkServer(); // Checks if a connection with the server is possible
       }
   } else {
       ksettings->checkEmbossDirectories();
       if (embassyHasChanged)
           mainKembossWindow->loadLists(PhylipnewCheckBox->checkState());
   }

   ksettings->writeSettings();
   // Saves the kemboss settings

   writeSettings();
   // Saves the settings of the configuration dialog (Position on screen, size, etc.)

   if (fontHasChanged)
      foreach (kembossWindow *window, mainKembossWindow->programWindows)
         qobject_cast<kprogramWindow *>(window)->setTextEditFonts();
   QDialog::accept();
}

void configDlg::helpRequested()
{
   QMessageBox::information(this, "kemboss", tr("Sorry, help is still missing here."));
}

void configDlg::readSettings()
{
    QSettings *settings;
    if (ksettings->portableApp()) {
        settings = new QSettings(ksettings->kembossPortableConfigDir()+"configDlg", QSettings::IniFormat);
    } else {
        settings = new QSettings(QSettings::UserScope, "kemboss.org", "configDlg");
    }
    QPoint pos = settings->value("conf_pos").toPoint();
    QSize size = settings->value("conf_size").toSize();
    if (!size.isEmpty()) resize(size);
    if (!pos.isNull()) move(pos);

    QSettings *kembossSettings;
    if (ksettings->portableApp()) {
        kembossSettings = new QSettings(ksettings->kembossPortableConfigDir()+"kemboss", QSettings::IniFormat);
    } else {
        kembossSettings = new QSettings(QSettings::UserScope, "kemboss.org", "kemboss");
    }
    PhylipnewCheckBox->setChecked(kembossSettings->value("phylipnew_installed", false).toBool());
    QString currentStyle = kembossSettings->value("style", "default").toString();
    int i = 0;
    while (i < StyleComboBox->count()) {
        if (StyleComboBox->itemText(i) == currentStyle) {
            StyleComboBox->setCurrentIndex(i);
            break;
        }
        i++;
    }
}

void configDlg::writeSettings()
{
    QSettings *settings;
    if (ksettings->portableApp()) {
        settings = new QSettings(ksettings->kembossPortableConfigDir()+"configDlg", QSettings::IniFormat);
    } else {
        settings = new QSettings(QSettings::UserScope, "kemboss.org", "configDlg");
    }
    settings->setValue("conf_pos", pos());
    settings->setValue("conf_size", size());

    QSettings *kembossSettings;
    if (ksettings->portableApp()) {
        kembossSettings = new QSettings(ksettings->kembossPortableConfigDir()+"kemboss", QSettings::IniFormat);
    } else {
        kembossSettings = new QSettings(QSettings::UserScope, "kemboss.org", "kemboss");
    }
    kembossSettings->setValue("phylipnew_installed", PhylipnewCheckBox->checkState());
}

void configDlg::on_StyleComboBox_activated(const QString &text)
{
   if (text != "default") {
      QApplication::setStyle(QStyleFactory::create(text));
      QApplication::setPalette(QApplication::style()->standardPalette());
   } else
      QMessageBox::information(this, "kemboss", "The default style will be used, when kemboss is restarted.");
   QSettings *settings;
   if (ksettings->portableApp()) {
       settings = new QSettings(ksettings->kembossPortableConfigDir()+"kemboss", QSettings::IniFormat);
   } else {
       settings = new QSettings(QSettings::UserScope, "kemboss.org", "kemboss");
   }
   settings->setValue("style", text);
}

void configDlg::on_OnlineDocCheckBox_stateChanged(int state)
{
    if (state == Qt::Checked) {
        EmbossHelpLabel->setEnabled(false);
        EmbossHelpLineEdit->setEnabled(false);
        EmbossHelpPushButton->setEnabled(false);
        OnlineDocSourceComboBox->setEnabled(true);
    } else {
        EmbossHelpLabel->setEnabled(true);
        EmbossHelpLineEdit->setEnabled(true);
        EmbossHelpPushButton->setEnabled(true);
        OnlineDocSourceComboBox->setEnabled(false);
    }
}

void configDlg::setExistingDir(const QString &dlgMsg, QLineEdit *lineEdit)
{
   QString dirName = QFileDialog::getExistingDirectory(this, dlgMsg, ksettings->currentDir(), QFileDialog::ShowDirsOnly | QFileDialog::DontResolveSymlinks);
   if (!dirName.isEmpty()) {
      lineEdit->setText(dirName);
      ksettings->setcurrentDir(dirName);
   }
}

void configDlg::setOpenFile(const QString &dlgMsg, QLineEdit *lineEdit)
{
   QString fileName = QFileDialog::getOpenFileName(this, dlgMsg, ksettings->currentDir());
   if (!fileName.isEmpty()) {
      lineEdit->setText(fileName);
      ksettings->setcurrentDir(QFileInfo(fileName).absolutePath());
   }
}

void configDlg::setSaveFile(const QString &dlgMsg, QLineEdit *lineEdit)
{
   QString fileName = QFileDialog::getSaveFileName(this, dlgMsg, ksettings->currentDir());
   if (!fileName.isEmpty()) {
      lineEdit->setText(fileName);
      ksettings->setcurrentDir(QFileInfo(fileName).absolutePath());
   }
}
