/***************************************************************************
 *                                                                         *
 *   kemboss - Graphical user interface for EMBOSS                         *
 *                                                                         *
 *   Copyright (C) 2007-2020 Tobias Schwarz (mail@tobias-schwarz.net)      *
 *                                                                         *
 *   Based on the EMBOSS.kaptn scripts by Thomas Siegmund                  *
 *                                                                         *
 *   This program is free software; you can redistribute and/or modify     *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version, and provided that the above       *
 *   copyright and permission notice is included with all distributed      *
 *   copies of this or derived software.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 **************************************************************************/

#include <QtCore/QCoreApplication>
#include <QTextStream>
#include <QSettings>
#include <QDir>
#include <QFile>
#include <QSslSocket>

#ifdef Q_OS_WIN32
    #include "windows.h"
#else
    // The Qt Cryptographic Architecture (QCA) from http://delta.affinix.com/qca/ has to be installed to be able to compile
    #include <QtCrypto>
#endif

#include "configure.h"
#include "../kembossd/sql.h"
#include "../kembossd/simplecrypt.h"

bool configureServer(const QString &configFile)
{
    QTextStream qstdout(stdout);
    QTextStream qstdin(stdin);
    QString input;
    QString def;
    uint intInput;
    QDir dir = QDir();
    QSettings settings(configFile, QSettings::IniFormat);

    qstdout << "Configuring server..." << endl;

    def = settings.value("Port", 5757).toString();
    qstdout << "TCP port to listen on (" << def << "): " << flush;
    input = qstdin.readLine();
    if (input == "") input = def;
    intInput = input.toUInt();
    settings.setValue("Port", intInput);

    def = settings.value("WelcomeMessage", "Welcome on kemboss server").toString();
    qstdout << "Welcome message (" << def << "): " << flush;
    input = qstdin.readLine();
    if (input == "") input = def;
    settings.setValue("WelcomeMessage", input);

    #ifdef Q_OS_WIN32
        def = settings.value("EmbossBinDir", "c:/Program files/mEMBOSS").toString();
    #else
        def = settings.value("EmbossBinDir", "/usr/bin/emboss").toString();
    #endif
    qstdout << "Emboss binary dir (" << def << "): " << flush;
    input = QDir::fromNativeSeparators(qstdin.readLine());
    if (input == "") input = def;
    if (!dir.exists(input)) {
        qstdout << "Emboss binary dir doesn't exist!" << endl;
    }
    settings.setValue("EmbossBinDir", input);

    #ifdef Q_OS_WIN32
        def = settings.value("EmbossProgramList", QCoreApplication::applicationDirPath() + "/allowed_emboss_programs").toString();
    #else
        def = settings.value("EmbossProgramList", "/etc/kembossd/allowed_emboss_programs").toString();
    #endif
    qstdout << "File containing the allowed EMBOSS programs (" << def << "): " << flush;
    input = QDir::fromNativeSeparators(qstdin.readLine());
    if (input == "") input = def;
    if (!QFile::exists(input)) {
        qstdout << "File doesn't exist! Please check Your settings and Your server installation." << endl;
    }
    settings.setValue("EmbossProgramList", input);

    if (settings.value("AcceptAllNewUsers", false).toBool()) def = "y"; else def = "n";
    qstdout << "Accept all new users: y/n (" << def << "): " << flush;
    input = qstdin.readLine();
    if (input == "") input = def;
    if (input == "y") settings.setValue("AcceptAllNewUsers", true); else settings.setValue("AcceptAllNewUsers", false);

    def = settings.value("UserDir", QDir::homePath()+"/kembossd/users").toString();
    qstdout << "Directory where the users input and output files are cached (" << def << "): " << flush;
    input = QDir::fromNativeSeparators(qstdin.readLine());
    if (input == "") input = def;
    settings.setValue("UserDir", input);

    def = settings.value("MaxNumberEmbossPrograms", 10).toString();
    qstdout << "Maximum number of EMBOSS programs executed in parallel (" << def << "): " << flush;
    input = qstdin.readLine();
    if (input == "") input = def;
    intInput = input.toUInt();
    settings.setValue("MaxNumberEmbossPrograms", intInput);

    def = settings.value("MaxEmbossProgramExecTime", 3600).toString();
    qstdout << "Maximum execution time of an EMBOSS program in seconds (" << def << "): " << flush;
    input = qstdin.readLine();
    if (input == "") input = def;
    intInput = input.toUInt();
    settings.setValue("MaxEmbossProgramExecTime", intInput);

    def = settings.value("Logfile", QDir::homePath()+"/kembossd/server.log").toString();
    qstdout << "File to log access to the server: either 'stdout' or path and filename (" << def << "): " << flush;
    input = QDir::fromNativeSeparators(qstdin.readLine());
    if (input == "") input = def;
    settings.setValue("Logfile", input);

    return(true);
}

bool configureSSL(const QString &configFile)
{
    QTextStream qstdout(stdout);
    QTextStream qstdin(stdin);
    QString input;
    QString def;

    QSettings settings(configFile, QSettings::IniFormat);

    qstdout << "Configuring SSL connection..." << endl;

    settings.beginGroup("SSL_connection");

    #ifdef Q_OS_WIN32
        def = settings.value("Privkey", QCoreApplication::applicationDirPath() + "/privkey.pem").toString();
    #else
        def = settings.value("Privkey", "/etc/kembossd/privkey.pem").toString();
    #endif
    qstdout << "Private key file (" << def << "): " << flush;
    input = qstdin.readLine();
    if (input == "") input = def;
    if (!QFile::exists(input)) {
        qstdout << "Private key file doesn't exist!" << endl;
    }
    settings.setValue("Privkey", input);

    #ifdef Q_OS_WIN32
        def = settings.value("Localcert", QCoreApplication::applicationDirPath() + "/localcert.pem").toString();
    #else
        def = settings.value("Localcert", "/etc/kembossd/localcert.pem").toString();
    #endif
    qstdout << "Local certificate file (" << def << "): " << flush;
    input = qstdin.readLine();
    if (input == "") input = def;
    if (!QFile::exists(input)) {
        qstdout << "Local certificate file doesn't exist!" << endl;
    }
    settings.setValue("Localcert", input);

    def = settings.value("Cacert", "none").toString();
    qstdout << "CA certificates path (" << def << "): " << flush;
    input = qstdin.readLine();
    if (input == "") input = def;
    if (!input.isEmpty() && input != "none") {
        QDir dir(input);
        if (!dir.exists()) {
            qstdout << "CA certificates path doesn't exist!" << endl;
        }
    }
    settings.setValue("Cacert", input);

    settings.endGroup();

    return true;
}

bool configureDb(const QString &configFile)
{
    QTextStream qstdout(stdout);
    QTextStream qstdin(stdin);
    QString input;
    QString def;
    QSettings settings(configFile, QSettings::IniFormat);

    qstdout << "Configuring SQL database..." << endl;

    settings.beginGroup("Database");

    def = settings.value("DatabaseType", "MYSQL").toString();
    qstdout << "Database type: MYSQL, SQLITE (" << def << "): " << flush;
    input = qstdin.readLine().toUpper();
    if (input == "") input = def;
    settings.setValue("DatabaseType", input);

    if (input == "MYSQL") {
        configureMYSQL(configFile);
    } else if (input == "SQLITE") {
        configureSQLITE(configFile);
    } else {
        qstdout << "Unknown database type!" << flush;
    }

    settings.endGroup();

    ksql::checkDb();

    return(true);
}

bool configureMYSQL(const QString &configFile)
{
    QTextStream qstdout(stdout);
    QTextStream qstdin(stdin);
    QString input;
    QString def;
    QSettings settings(configFile, QSettings::IniFormat);

    settings.beginGroup("MySQL");

    def = settings.value("DatabaseHostname", "localhost").toString();
    qstdout << "Database hostname (" << def << "): " << flush;
    input = qstdin.readLine();
    if (input == "") input = def;
    settings.setValue("DatabaseHostname", input);

    def = settings.value("DatabasePort", "3306").toString();
    qstdout << "Database port (" << def << "): " << flush;
    input = qstdin.readLine();
    if (input == "") input = def;
    settings.setValue("DatabasePort", input);

    def = settings.value("DatabaseName", "kemboss").toString();
    qstdout << "Database name (" << def << "): " << flush;
    input = qstdin.readLine();
    if (input == "") input = def;
    settings.setValue("DatabaseName", input);

    def = settings.value("DatabaseUser", "kemboss").toString();
    qstdout << "Database user (" << def << "): " << flush;
    input = qstdin.readLine();
    if (input == "") input = def;
    settings.setValue("DatabaseUser", input);

#ifdef Q_OS_WIN32
    input = "";
    do {
        qstdout << "Database password: " << flush;

        enableEcho(false);
        input = qstdin.readLine();
        enableEcho(true);

        qstdout << endl;

    } while (input == ""); // Empty password is not allowed
#else
    QCA::SecureArray pass;
    QCA::ConsolePrompt prompt;
    do {
        prompt.getHidden("Database password: ");
        prompt.waitForFinished();
        pass = prompt.result();
    } while (pass.isEmpty()); // Empty password is not allowed
#endif

    SimpleCrypt crypto(Q_UINT64_C(0x1c3ad4a3acb9f023));

#ifdef Q_OS_WIN32
    settings.setValue("DatabasePassword", crypto.encryptToString(input));
#else
    settings.setValue("DatabasePassword", crypto.encryptToString(pass.toByteArray()));
#endif

    settings.endGroup();

    return true;
}

bool configureSQLITE(const QString &configFile)
{
    QTextStream qstdout(stdout);
    QTextStream qstdin(stdin);
    QString input;
    QString def;
    QSettings settings(configFile, QSettings::IniFormat);

    settings.beginGroup("SQLite");

    def = settings.value("DatabaseName", QDir::homePath()+"/kembossd/database.sqlite").toString();
    qstdout << "Database file (" << def << "): " << flush;
    input = QDir::fromNativeSeparators(qstdin.readLine());
    if (input == "") input = def;
    settings.setValue("DatabaseName", input);

    // Check if database dir exists
    QFileInfo databaseFile(input);
    QDir databasePath(databaseFile.absolutePath());
    if (!databasePath.exists()) {
        if (databasePath.mkpath(databaseFile.absolutePath()))
            qstdout << "Created database dir " << databaseFile.absolutePath() << endl;
        else
            qstdout << "Creating database dir failed. Please review Your settings." << endl;
    }

    settings.endGroup();

    return true;
}

bool configureEmbassyPackages(const QString &configFile)
{
    QTextStream qstdout(stdout);
    QTextStream qstdin(stdin);
    QString input;
    QString def;
    QSettings settings(configFile, QSettings::IniFormat);

    settings.beginGroup("EMBASSY_packages");

    qstdout << "Select installed EMBASSY packages..." << endl << endl;

    if (settings.value("PHYLIPNEW", true).toBool()) def = "y"; else def = "n";
    qstdout << "PHYLIPNEW installed: y/n (" << def << "): " << flush;
    input = qstdin.readLine();
    if (input == "") input = def;
    if (input == "y") settings.setValue("PHYLIPNEW", true); else settings.setValue("PHYLIPNEW", false);

    settings.endGroup();

    return true;
}


#ifdef Q_OS_WIN32
void enableEcho(bool enable)
{
    HANDLE stdinHandle = GetStdHandle(STD_INPUT_HANDLE);
    DWORD  mode;

    GetConsoleMode(stdinHandle, &mode);

    if (!enable)
    {
        mode &= ~ENABLE_ECHO_INPUT;
    }
    else
    {
        mode |= ENABLE_ECHO_INPUT;
    }

    SetConsoleMode(stdinHandle, mode);
}
#endif
