/***************************************************************************
 *                                                                         *
 *   kemboss - Graphical user interface for EMBOSS                         *
 *                                                                         *
 *   Copyright (C) 2007-2020 Tobias Schwarz (mail@tobias-schwarz.net)      *
 *                                                                         *
 *   Based on the EMBOSS.kaptn scripts by Thomas Siegmund                  *
 *                                                                         *
 *   This program is free software; you can redistribute and/or modify     *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version, and provided that the above       *
 *   copyright and permission notice is included with all distributed      *
 *   copies of this or derived software.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 **************************************************************************/

#include "kthread.h"
#include "simplecrypt.h"
#include "sql.h"
#include "kserver.h"

#include <QtNetwork>
#include <QMetaType>
#include <QDebug>

KembossThread::KembossThread(qintptr socketDescriptor, QObject *parentServer)
    : socketDescriptor(socketDescriptor)
{
    qDebug() << "Creating connection object";

    server = parentServer;

    inputFilename = "";
    blockSize = 0;
    userID = 0;
}

KembossThread::~KembossThread()
{
    qDebug() << "Deleting connection object";
    timer->deleteLater();
    programProcess->deleteLater();
    tcpSocket->deleteLater();
}

void KembossThread::run()
{
    qDebug() << "Run server thread";

    setThreadStatus(kembossServer::ThreadStartingUp);

    programProcess = new QProcess(nullptr);

    timer = new QTimer(nullptr);
    connect(timer, SIGNAL(timeout()), this, SLOT(timeout()));
    timer->start(60000);
    lastDataTransfer.start();
    runTime.start();

    // Create a TCP socket and make it use our accepted socket
    tcpSocket = new QSslSocket(nullptr);

    tcpSocket->setProtocol(QSsl::SecureProtocols);
    if (!tcpSocket->setSocketDescriptor(socketDescriptor)) {
        emit error(tcpSocket->error());
        logEmbossRun(kembossServer::ClientServerConnectionFailed);
        connectionClosed();
        return;
    }

    // Connect the TCP socket's signals to our slots
    qRegisterMetaType<QAbstractSocket::SocketState>("SocketState");
    connect(tcpSocket, SIGNAL(stateChanged(QAbstractSocket::SocketState)), this, SLOT(stateChanged(QAbstractSocket::SocketState)));
    connect(tcpSocket, SIGNAL(encrypted()), this, SLOT(encryptedConnection()));
    connect(tcpSocket, SIGNAL(proxyAuthenticationRequired(const QNetworkProxy &, QAuthenticator *)), this, SLOT(proxyAuthenticationRequired(const QNetworkProxy &, QAuthenticator *)));
    qRegisterMetaType<QAbstractSocket::SocketError>("SocketError");
    connect(tcpSocket, SIGNAL(error(QAbstractSocket::SocketError)), this, SLOT(tcpError(QAbstractSocket::SocketError)));
    qRegisterMetaType<QSslError>("QSslError");
    connect(tcpSocket, SIGNAL(peerVerifyError(const QSslError &)), this, SLOT(peerVerifyError(const QSslError &)));
    qRegisterMetaType<QList<QSslError> >("QListQSslError");
    connect(tcpSocket, SIGNAL(sslErrors(const QList<QSslError> &)), this, SLOT(sslErrors(const QList<QSslError> &)));
    connect(tcpSocket, SIGNAL(readyRead()), this, SLOT(readData()));
    connect(tcpSocket, SIGNAL(disconnected()), this, SLOT(connectionClosed()));

    msgStdOut("Accepted client " + tcpSocket->peerAddress().toString());

    setThreadStatus(kembossServer::TransferringConfiguration);
    // Upon successfull connection, the kemboss server sends the current server status to the client
    startServerEncryption();
    //exec(); An event loop is started here by the thread containing the connection object
}

void KembossThread::stateChanged(QAbstractSocket::SocketState socketState)
{
    switch (socketState) {
        case QAbstractSocket::UnconnectedState: qDebug() << "UnconnectedState";
            break;
        case QAbstractSocket::HostLookupState: qDebug() << "HostLookupState";
            break;
        case QAbstractSocket::ConnectingState: qDebug() << "ConnectingState";
            break;
        case QAbstractSocket::ConnectedState: qDebug() << "ConnectedState";
            break;
        case QAbstractSocket::BoundState: qDebug() << "BoundState";
            break;
        case QAbstractSocket::ClosingState: qDebug() << "ClosingState";
            break;
        case QAbstractSocket::ListeningState: qDebug() << "ListeningState";
            break;
    }
}

bool KembossThread::startServerEncryption()
{
    qDebug() << "Starting server encryption";
    msgStdOut(tcpSocket->peerAddress().toString()+": Starting SSL handshake");

    // We do not want to verify the identity of the client
    tcpSocket->setPeerVerifyMode(QSslSocket::VerifyNone);

    // Trying to open the private key file
    QFile file(qobject_cast<KembossServer *>(server)->getPrivateKeyFile());
    file.open(QIODevice::ReadOnly);
    if (!file.isOpen()) {
        msgStdOut(tcpSocket->peerAddress().toString()+": Couldn't open private key file" + file.fileName());
        return false;
    }

    QSslKey key(&file, QSsl::Rsa, QSsl::Pem, QSsl::PrivateKey);
    if (key.isNull()) {
        msgStdOut(tcpSocket->peerAddress().toString()+": Private key is invalid");
        return false;
    }
    file.close();

    // We can add some CA certificates if we want to
    QString path = qobject_cast<KembossServer *>(server)->getCACertsPath();
    if (!path.isEmpty() && !tcpSocket->addCaCertificates(path, QSsl::Pem, QRegExp::Wildcard))
    {
        msgStdOut(tcpSocket->peerAddress().toString()+": Couldn't add CA certificates");
    }

    // Then we set the local certificate
    file.setFileName(qobject_cast<KembossServer *>(server)->getLocalCertFile());
    if (!file.open(QIODevice::ReadOnly)) {
        msgStdOut(tcpSocket->peerAddress().toString()+": Couldn't open local certificate file" + file.fileName());
        return false;
    }

    QSslCertificate cert(&file, QSsl::Pem);
    if (cert.isNull()) {
        msgStdOut(tcpSocket->peerAddress().toString()+": Local certificate isn't valid!");
        return false;
    }
    if (cert.isBlacklisted()) {
        msgStdOut(tcpSocket->peerAddress().toString()+": Local certificate isn't valid!");
        return false;
    }
    QDateTime currDateTime = QDateTime::currentDateTimeUtc();
    if (cert.effectiveDate() > currDateTime || cert.expiryDate() < currDateTime) {
        msgStdOut(tcpSocket->peerAddress().toString()+": Local certificate has expired!");
        return false;
    }

    // We set the local certificate and the private key and start server encryption
    tcpSocket->setLocalCertificate(cert);
    tcpSocket->setPrivateKey(key);
    tcpSocket->startServerEncryption();
    return true;
}

void KembossThread::encryptedConnection()
{
    qDebug() << "Encrypted connection";
    msgStdOut(tcpSocket->peerAddress().toString()+": Encrypted connection established");
    // after server encryption is completed the server sends its status to the client
    writeServerStatus();
}

void KembossThread::readData()
{
    // If no EMBOSS process is running and no data is transferred for 60 secs, the thread is terminated
    lastDataTransfer.restart();

    // First, read all incoming data from the client.
    QDataStream in(tcpSocket);
    in.setVersion(QDataStream::Qt_4_6);

    // The client response starts with the size of the packet, so first we need to ensure that we can read the size
    if (blockSize == 0) {
        if (tcpSocket->bytesAvailable() < int(sizeof(qint64)))
            return;
        in >> blockSize;
    }

    // Then we will wait until the socket has received the complete command of the client
    if (tcpSocket->bytesAvailable() < blockSize)
        return;

    // If we have received the complete command of the client, it is read into a QByteArray
    QByteArray incoming;
    in >> incoming;

    qDebug() << "Incoming data:";

    // The first 5 characters of the incoming data are the command.
    QByteArray command = incoming.left(5);
    incoming.remove(0, 5);

    if (userID == 0) {
        // As long as the user has not been verified, the only accepted command is the user identification
        if (command == "[cvn]") { // client version number followed by the user name, user group, user email and user password hash.
            userID = readUser(incoming);
        } else {
            msgStdOut(tcpSocket->peerAddress().toString()+": Received unknown command. Closing connection");
            logEmbossRun(kembossServer::ServerReceivedUnknownCommand);
            closeConnection();
        }
    } else if (userID > 0) {
        // After user identification the other commands are accepted
               if (command == "[epn]") { // EMBOSS program name followed by the EMBOSS program configuration
            readEmbossConfiguration(QString::fromUtf8(incoming));
        } else if (command == "[eol]") { // EMBOSS output file list
            readEmbossOutputFileList(QString::fromUtf8(incoming));
        } else if (command == "[fid]") { // EMBOSS input filename (without path) followed by the md5 sum of the file
            readEmbossInputFileID(QString::fromUtf8(incoming));
        } else if (command == "[eif]") { // EMBOSS input file
            readEmbossInputFile(incoming);
        } else if (command == "[rap]") { // transfer of input files finished, run EMBOSS app
            runEmbossApp();
        } else if (command == "[ore]") { // EMBOSS output / output file received, send next EMBOSS output file identifier
            writeEmbossOutputFileID();
        } else if (command == "[oir]") { // EMBOSS output file identifier received, send EMBOSS output file
            writeEmbossOutputFile(serverEmbossOutputFiles.takeFirst());
        } else if (command == "[pwu]") { // new user password
            readNewPassword(incoming);
        } else if (command == "[cpe]") { // Execution canceled by user
            userCanceled();
        } else {
            msgStdOut(tcpSocket->peerAddress().toString()+": Received unknown command. Closing connection");
            logEmbossRun(kembossServer::ServerReceivedUnknownCommand);
            closeConnection();
        }
    }

    // The blockSize is reset to 0 so that the next command can be retrieved from the client
    blockSize = 0;
}

qint64 KembossThread::writeData(QByteArray data)
{
    lastDataTransfer.restart();
    // If no EMBOSS process is running and no data is transferred for 60 secs, the thread is terminated

    qDebug() << "Writing data to network:";

    QByteArray block;
    QDataStream out(&block, QIODevice::WriteOnly);
    out.setVersion(QDataStream::Qt_4_6);
    out << qint64(0);
    out << data;
    out.device()->seek(0);
    out << qint64(block.size() - sizeof(qint64));
    return tcpSocket->write(block);
}

void KembossThread::writeServerStatus()
{
    qDebug() << "Writing server status";
    QString sst;
    sst.setNum(qobject_cast<KembossServer *>(server)->getServerStatus());
    writeData(QString("[sst]"+sst+"[spl]"+qobject_cast<KembossServer *>(server)->getWelcomeMessage()).toUtf8());
}

quint32 KembossThread::readUser(const QByteArray &data)
{
    qDebug() << "Receiving client version and user";
    msgStdOut(tcpSocket->peerAddress().toString()+": Receiving client version and user");

    // The client version, user name, user group, user email and user password hash are separated by the characters [spl]
    QStringList dataList = QString(data).split("[spl]");

    clientVersion = dataList.at(0).toUShort();

    QString userName = dataList.at(1);
    QString userGroup = dataList.at(2);
    QString userEmail = dataList.at(3);

    QByteArray pwHash = kembossServer::SHA256hash(dataList.at(4).toLatin1());
    // A single ' in the password hash has to be escaped by '', otherwise the SQL statement checking the user is invalid
    pwHash.replace("'","''");

    if (clientVersion >= 210)
        clientOS = dataList.at(5).toUShort();
    else
        clientOS = 0;

    quint32 nameID = 0;
    quint32 uid = 0;
    QString idStr;
    bool updatePw = false;

    // Check if userName exists
    nameID = ksql::execSqlQuery("SELECT id FROM user WHERE name = '"+userName+"'",
                                tcpSocket->peerAddress().toString()+": Error in verifying the user data.", true).toUInt();

    if (nameID == 0) {
        qDebug() << "User doesn't exist";
        // If the userName doesn't exist in the database, we have to check if we accept all new users
        if (qobject_cast<KembossServer *>(server)->getAcceptAllUsers()) {
            // If the user does not exist in the database but all new users are accepted, the new user is stored in the database
            qDebug() << "Save new user to database";
            uid = ksql::newUser(userName, userGroup, pwHash, false, userEmail);
            if (uid == 0) {
                msgStdErr(tcpSocket->peerAddress().toString()+": Error in saving new user. Exiting.");
                userAccessDenied(false);
                return 0;
            }
            idStr.setNum(uid);
            msgStdOut(tcpSocket->peerAddress().toString()+": New user " + userName + " created, user ID: " + idStr);
        } else {
            userAccessDenied(false);
            return 0;
        }
    } else {
        // If the userName exists in the database we have to check if the password is correct
        uid = ksql::execSqlQuery("SELECT id FROM user WHERE name = '"+userName+"' AND pw = '"+pwHash+"'",
                                 tcpSocket->peerAddress().toString()+": Error in verifying the user data.", true).toUInt();
        if (uid > 0) {
            // If the user exists in the database, the field 'last_login' in the database is updated
            idStr.setNum(uid);
            ksql::execSqlQuery("UPDATE user SET last_login='"+QDateTime::currentDateTime().toString("yyyy-MM-dd hh:mm:ss")+"' WHERE id="+idStr,
                               tcpSocket->peerAddress().toString()+": Error in updating the 'last_login' of the user database.");
            msgStdOut(tcpSocket->peerAddress().toString()+": User " + userName + " accepted, user ID: " + idStr);

            // We have to check if the user has to update his password
            updatePw = ksql::execSqlQuery("SELECT pw_change_required FROM user WHERE id="+idStr,
                                          tcpSocket->peerAddress().toString()+": Error in verifying if an password update is required.", true).toBool();
        } else {
            // The username exists in the database but the password is incorrect
            userAccessDenied(true);
            return 0;
        }
    }

    // We create the input and output directory here, as we have the user name at hand
    serverInputDir.setPath(qobject_cast<KembossServer *>(server)->getEmbossFileDir() + userName + "/input_files");
    serverOutputDir.setPath(qobject_cast<KembossServer *>(server)->getEmbossFileDir() + userName + "/" + QDateTime::currentDateTime().toString("yyyy-MM-dd hh-mm-ss"));

    // If the user is accepted, send an acknowledgment to the client
    if (updatePw) {
        writeData("[npw]"); // Send user acknowledged, password update required
    } else {
        writeData("[uak]"); // Send user acknowledged
    }

    return uid; // The function returns the user ID or 0 if user identification failed
}

void KembossThread::userAccessDenied(const bool usernameFound)
{
    qDebug() << "User access denied";
    if (usernameFound) {
        writeData("[uwp]"); // username was found in the database but the entered password was wrong
    } else {
        writeData("[ude]"); // Send user denied
    }
    msgStdOut(tcpSocket->peerAddress().toString()+": User access denied.");
    logEmbossRun(kembossServer::UserIdentificationFailed);
    closeConnection();
}

void KembossThread::readNewPassword(const QByteArray &data)
{
    qDebug() << "Receiving new password";
    msgStdOut(tcpSocket->peerAddress().toString()+": Updating user password.");

    QString idStr;
    idStr.setNum(userID);

    QByteArray pwHash = kembossServer::SHA256hash(data);

    ksql::execSqlQuery("UPDATE user SET pw='"+pwHash+"', pw_change_required=FALSE WHERE id="+idStr,
                       tcpSocket->peerAddress().toString()+": Error in updating the password.");

    writeData(QString("[uak]").toUtf8()); // Send user acknowledged
}

void KembossThread::readEmbossConfiguration(const QString &data)
{
    qDebug() << "Receiving EMBOSS configuration";
    msgStdOut(tcpSocket->peerAddress().toString()+": Receiving EMBOSS configuration");
    programConfiguration = data.split("[spl]", QString::SkipEmptyParts);
    embossApp = programConfiguration.takeFirst();
    if (embossApp == "checkServer") {
        checkServer();
    } else
        if (checkEmbossApp()) {
            msgStdOut(tcpSocket->peerAddress().toString()+": Execution of "+embossApp+" allowed");
            writeData(QString("[eea]").toUtf8()); // Execution of EMBOSS binary allowed. Send EMBOSS output file list
        } else {
            closeConnection();
        }
}

void KembossThread::readEmbossOutputFileList(const QString &data)
{
    qDebug() << "Receiving EMBOSS output file list";
    msgStdOut(tcpSocket->peerAddress().toString()+": Receiving EMBOSS output file list");
    clientEmbossOutputFiles = data.split("[spl]", QString::SkipEmptyParts);

    // The embossOutputFiles list contains the output files including the path to the file on the client
    // Within the programConfiguration the paths to the output files are replaced with the output paths on the server
    for (int i = 0; i < clientEmbossOutputFiles.size(); ++i) {
        QFileInfo fileinfo(clientEmbossOutputFiles.at(i));
        programConfiguration.replaceInStrings(clientEmbossOutputFiles.at(i), serverOutputDir.path()+"/"+fileinfo.fileName(), Qt::CaseInsensitive);
    }
    writeData(QString("[eor]").toUtf8()); // Output file list received, send EMBOSS input file identifier
}

void KembossThread::readEmbossInputFileID(const QString &data)
{
    qDebug() << "Receiving EMBOSS input file id";
    msgStdOut(tcpSocket->peerAddress().toString()+": Receiving EMBOSS input file identifier");

    // The input filename including the path on the client is received
    clientInputFile = data.section("[spl]",0,0);

    // The md5sum of the input file is received from the client
    inputMD5sum = data.section("[spl]",1,1);

    // Check if a file with the received md5sum already exists in the filecache.
    inputFilename = checkIfFileIsCached(inputMD5sum);

    if (inputFilename == "0") {

        // The file is not cached on the server and has to be retrieved
        QFileInfo fileinfo(clientInputFile);
        inputFilename = fileinfo.fileName();
        msgStdOut(tcpSocket->peerAddress().toString()+": Input file " + inputFilename + " is not cached, has to be transferred");

        // The input filename(s) and path(s) in the EMBOSS configuration are changed after the file transfer has occured, as the filename might change during file transfer

        writeData("[get]");
    } else {

        // The file is already cached on the server. The filename of the input file is the filename of the cached file
        msgStdOut(tcpSocket->peerAddress().toString()+": Input file " + inputFilename + " is already cached on the server");

        // The input filename(s) and path(s) in the EMBOSS configuration have to be changed
        programConfiguration.replaceInStrings(clientInputFile, serverInputDir.path()+"/"+inputFilename, Qt::CaseInsensitive);

        writeData("[has]");
    }
}

QString KembossThread::checkIfFileIsCached(const QString &md5sum)
{
    qDebug() << "Checking if file is cached";
    QString userIDStr;
    userIDStr.setNum(userID);
    QString filename = ksql::execSqlQuery("SELECT file FROM input_files WHERE md5 = '"+md5sum+"' AND user = '"+userIDStr+"'",
                              tcpSocket->peerAddress().toString()+": Error in checking file cache", true).toString();
    if (filename != "0") {
        // If a file with the given md5sum is stored in the SQL database, it is checked if the file really exists in the input file cache
        if (!QFile::exists(serverInputDir.path()+"/"+filename)) {
            // If a file is stored in the SQL database but is not present in the input file cache, the SQL database is updated and an error message is shown
            msgStdErr(tcpSocket->peerAddress().toString()+": An input file is stored in the SQL database but is not present in the input file cache. Updating database");
            ksql::execSqlQuery("DELETE FROM input_files WHERE md5 = '"+md5sum+"' AND user = '"+userIDStr+"'",
                               tcpSocket->peerAddress().toString()+": Error in updating file cache");
            return "0";
        }
        // If a file with the given md5sum really exists in the input file cache, the last_accessed field in the database is updated
        ksql::execSqlQuery("UPDATE input_files SET last_accessed='"+QDateTime::currentDateTime().toString("yyyy-MM-dd hh:mm:ss")+"' WHERE md5 = '"+md5sum+"' AND user = '"+userIDStr+"'",
                           tcpSocket->peerAddress().toString()+": Error in updating the 'last_accessed' of the input_files database.");
    }

    return filename;
}

void KembossThread::readEmbossInputFile(const QByteArray &data)
{
    qDebug() << "Receiving EMBOSS input file";
    setThreadStatus(kembossServer::TransferringFile);
    msgStdOut(tcpSocket->peerAddress().toString()+": Receiving EMBOSS input file " + inputFilename);

    // First we have to check if a file with the inputFilename already exists in the filecache
    quint16 fileNum = 0;
    QString origFilename = inputFilename;
    while (QFile::exists(serverInputDir.path()+"/"+inputFilename)) {
        // If a file with the inputFilename already exists, a number is appended to the filename until the file doesn't exist
        fileNum++;
        QString fileNumStr;
        fileNumStr.setNum(fileNum);
        inputFilename = origFilename + "." + fileNumStr;
    }

    // The input file is received, uncompressed and saved in the inputDir
    QFile file(serverInputDir.path()+"/"+inputFilename);
    if (!file.open(QIODevice::WriteOnly)) {
        msgStdOut(tcpSocket->peerAddress().toString()+": The input file " + inputFilename + " could not be saved!");
        logEmbossRun(kembossServer::FileTransferError);
        closeConnection();
        return;
    }
    file.write(qUncompress(data));
    file.close();

    // Following the file transfer the md5 sum of the transferred file is compared to the md5 sum sent by the client
    if (QString::fromUtf8(kemboss::getFileMD5Sum(serverInputDir.path()+"/"+inputFilename)) == inputMD5sum ) {
        // If the MD5sums are identical the file information is stored in the SQL database
        QString timestampStr = QDateTime::currentDateTime().toString("yyyy-MM-dd hh:mm:ss");
        QString userIDStr;
        userIDStr.setNum(userID);
        ksql::execSqlQuery("INSERT INTO input_files (`user`, `file`, `md5`, `last_accessed`) "
                           "VALUES ('"+userIDStr+"', '"+inputFilename+"', '"+inputMD5sum+"', '"+timestampStr+"')",
                           "Error in writing the input file information to the database.");
    } else {
        // If the MD5sums are not identical some error occurred during file transfer. The received file is deleted and the connection is terminated
        // In an advanced version of the server the file transfer could be repeated
        msgStdErr(tcpSocket->peerAddress().toString()+": An error occured during transfer of the input file " + inputFilename);
        QFile::remove(serverInputDir.path()+"/"+inputFilename);
        logEmbossRun(kembossServer::FileTransferError);
        closeConnection();
        return;
    }

    // The input filename(s) and path(s) in the EMBOSS configuration have to be changed
    programConfiguration.replaceInStrings(clientInputFile, serverInputDir.path()+"/"+inputFilename, Qt::CaseInsensitive);

    // The input file variables are cleared after transfer of the file is finished
    clientInputFile = "";
    inputFilename = "";
    inputMD5sum = "";

    setThreadStatus(kembossServer::TransferringConfiguration);
    writeData(QString("[fin]").toUtf8()); // Send next EMBOSS input file identifier
}

bool KembossThread::checkEmbossApp()
{
    qDebug() << "Checking EMBOSS app";
    // To avoid conflicting binary names, on some systems chaos is renamed into em_chaos
    if (embossApp == "chaos" && QFile::exists(qobject_cast<KembossServer *>(server)->getEmbossBinDir()+"em_chaos"))
        embossApp = "em_chaos";

    // Check if execution of program is allowed (program is listed in the file allowed_emboss_programs)
    if (!qobject_cast<KembossServer *>(server)->embossProgramExecutionAllowed(embossApp)) {
        msgStdOut(tcpSocket->peerAddress().toString()+": Execution of program not allowed. Exiting");
        writeData(QString("[pna]").toUtf8());
        logEmbossRun(kembossServer::ExecutionOfProgramNotAllowed);
        return false;
    }

    // Check if binary exists
    if (!QFile::exists(qobject_cast<KembossServer *>(server)->getEmbossBinDir()+embossApp
    #ifdef Q_OS_WIN32
        +".exe"
    #endif
    )) {
        msgStdOut(tcpSocket->peerAddress().toString()+": Emboss binary " + qobject_cast<KembossServer *>(server)->getEmbossBinDir()+embossApp + " not found. Exiting");
        writeData(QString("[bnf]").toUtf8());
        logEmbossRun(kembossServer::EmbossBinaryNotFound);
        return false;
    }

    // If execution of the EMBOSS program is allowed, we create the input and output file dirs
    if (!serverInputDir.exists()) {
        if (!serverInputDir.mkpath(serverInputDir.path())) {
            msgStdErr(tcpSocket->peerAddress().toString()+": Could not create directory for users input files");
            logEmbossRun(kembossServer::FileTransferError);
            return false;
        }
    }
    serverOutputDir.setPath(serverOutputDir.path() + " " + embossApp);
    if (!serverOutputDir.mkpath(serverOutputDir.path())) {
        msgStdErr(tcpSocket->peerAddress().toString()+": Could not create directory for users output files");
        logEmbossRun(kembossServer::FileTransferError);
        return false;
    }

    return true;
}

void KembossThread::runEmbossApp()
{
    setThreadStatus(kembossServer::EmbossProgramRunning);
    msgStdOut(tcpSocket->peerAddress().toString() + ": Running EMBOSS app "+embossApp);

    // The EMBOSS process is allowed to run for a maximum of cmaxEmbossProcessTime secs
    embossProcessRuntime.start();

    // Standard output as well as standard error data are both shown
    programProcess->setProcessChannelMode(QProcess::MergedChannels);
    connect(programProcess, SIGNAL(readyReadStandardOutput()), this, SLOT(readProgramOutput()));
    connect(programProcess, SIGNAL(finished(int)), this, SLOT(embossProcessFinished(int)));
    connect(programProcess, SIGNAL(error(QProcess::ProcessError)), this, SLOT(embossProcessError(QProcess::ProcessError)));
    QString program = qobject_cast<KembossServer *>(server)->getEmbossBinDir()+embossApp;
    programProcess->start(program
    #ifdef Q_OS_WIN32
        +".exe"
    #endif
    , programConfiguration, QIODevice::ReadWrite);
}

void KembossThread::readProgramOutput()
{
    qDebug() << "Reading program output";
    cprogramOutput.append(programProcess->readAll());
}

void KembossThread::embossProcessFinished(int exitCode)
{
    setThreadStatus(kembossServer::TransferringEmbossOutput);
    QString num;
    num.setNum(exitCode);
    msgStdOut(tcpSocket->peerAddress().toString()+": EMBOSS program finished. Exit code is " + num);

    // Load all files in the EMBOSS output dir in serverEmbossOutputFiles for transfer to the client
    serverEmbossOutputFiles = serverOutputDir.entryList(QDir::Files);

    writeEmbossOutput();

    if (exitCode != 0) {
        logEmbossRun(kembossServer::UnknownErrorOnEmbossRun);
        closeConnection();
    }
}

void KembossThread::writeEmbossOutput()
{
    qDebug() << "Writing EMBOSS output";
    if (!cprogramOutput.isEmpty()) {
        // Save EMBOSS output to disk
        QFile file(serverOutputDir.path()+"/embossout.txt");
        if (!file.open(QIODevice::WriteOnly | QIODevice::Text)) {
            msgStdErr(tcpSocket->peerAddress().toString()+": Error in writing the EMBOSS output to disk");
        } else {
            QTextStream out(&file);
            out << cprogramOutput;
            file.close();
        }
    }
    writeData(QString("[eou]"+QString(cprogramOutput)).toUtf8());
}

void KembossThread::writeEmbossOutputFileID()
{
    qDebug() << "Writing EMBOSS output file ID";
    if (serverEmbossOutputFiles.isEmpty()) {
        fileTransferFinished();
    } else {
        writeData("[fod]" + serverEmbossOutputFiles.first().toUtf8() + "[spl]" + kemboss::getFileMD5Sum(serverOutputDir.absolutePath()+"/"+serverEmbossOutputFiles.first()));
    }
}

bool KembossThread::writeEmbossOutputFile(const QString &filename)
{
    qDebug() << "Writing EMBOSS output file";
    msgStdOut(tcpSocket->peerAddress().toString()+": Sending EMBOSS output file " + filename);

    QFile file(serverOutputDir.absolutePath()+"/"+filename);
    if (!file.open(QIODevice::ReadOnly)) {
        msgStdErr(tcpSocket->peerAddress().toString()+": Could not open EMBOSS output file " + filename + " for transfer to the client.");
        writeEmbossOutputFileID();
        return false;
    } else {
        writeData("[eof]" + qCompress(file.readAll()));
        file.close();
        return true;
    }
}

void KembossThread::fileTransferFinished()
{
    qDebug() << "File transfer finished";
    msgStdOut(tcpSocket->peerAddress().toString()+": File transfer finished. EMBOSS execution succeeded");
    logEmbossRun(kembossServer::Success);
    writeData("[tof]"); // transfer of output files finished, closing connection
    closeConnection();
}

void KembossThread::checkServer()
{
    qDebug() << "Server check by client";
    msgStdOut(tcpSocket->peerAddress().toString()+": Server check by client");
    QByteArray command = "[iep]";
    if (qobject_cast<KembossServer *>(server)->getPhylipnewInstalled() == true) command += "p1";  else  command += "p0";
    writeData(command); // Send installed EMBASSY packages to the client
    logEmbossRun(kembossServer::CheckServerSuccess);
    closeConnection();
}

void KembossThread::userCanceled()
{
    qDebug() << "User canceled";
    msgStdOut(tcpSocket->peerAddress().toString()+": User canceled execution of EMBOSS program. Terminating");
    killEmbossProcess();
    logEmbossRun(kembossServer::CanceledByUser);
}

void KembossThread::timeout()
{
    qDebug() << "Timeout";
    if (programProcess->state() == QProcess::NotRunning) {
        if (lastDataTransfer.elapsed() > 60000)
        {
            // If no EMBOSS process is running and no data has been transferred for more than 60 secs, the thread is terminated
            msgStdOut(tcpSocket->peerAddress().toString()+": No data was transferred for more than 60 secs. Terminating.");
            logEmbossRun(kembossServer::DataTransferTimeout);
            closeConnection();
        }
    } else {
        if (embossProcessRuntime.elapsed() > qobject_cast<KembossServer *>(server)->getMaxEmbossProcessTime()*1000) {
            // If an EMBOSS process is running and the process runtime (in msecs) exceeds the maximum allowed time (in secs), the thread is terminated
            msgStdOut(tcpSocket->peerAddress().toString()+": Maximum EMBOSS program runtime exceeded. Terminating");
            killEmbossProcess();
            logEmbossRun(kembossServer::MaxExecTimeExceeded);
            closeConnection();
        }
    }
}

void KembossThread::killEmbossProcess()
{
    qDebug() << "Killing EMBOSS process";
    if (programProcess->state() != QProcess::NotRunning) programProcess->kill();
    // Wird bei 'kill' das Signal 'embossProcessFinished' aufgerufen????
    closeConnection();
}

void KembossThread::logEmbossRun(const kembossServer::serverReturnCode returnCode)
{
    qDebug() << "Logging EMBOSS run";
    QString userIDStr;
    userIDStr.setNum(userID);
    QString timeElapsed;
    timeElapsed.setNum(runTime.elapsed());
    QString returnCodeStr;
    returnCodeStr.setNum(returnCode);
    QString clientVersionStr;
    clientVersionStr.setNum(clientVersion);
    QString clientOSStr;
    clientOSStr.setNum(clientOS);

    ksql::execSqlQuery("INSERT INTO program_run (user_id, ip, program, runtime, success, client_version, client_os)"
                       "VALUES ("+userIDStr+", '"+tcpSocket->peerAddress().toString()+"' ,'"+embossApp+"', "+timeElapsed+" ,"+returnCodeStr+" ,"+clientVersionStr+" ,"+clientOSStr+")",
                       tcpSocket->peerAddress().toString()+": Error in writing the emboss run data to the database.");

}

void KembossThread::closeConnection()
{
    qDebug() << "Closing connection";
    setThreadStatus(kembossServer::ClosingConnection);
    tcpSocket->disconnectFromHost();
}

void KembossThread::abortConnection()
{
    qDebug() << "Aborting connection";
    setThreadStatus(kembossServer::ClosingConnection);
    tcpSocket->abort();
}

void KembossThread::connectionClosed()
{
    msgStdOut(tcpSocket->peerAddress().toString()+": Connection closed");
    threadFinished(this->thread()); // Tells the server that the thread is finished. The server deletes the thread
}

void KembossThread::proxyAuthenticationRequired(const QNetworkProxy &, QAuthenticator *)
{
    msgStdErr(tcpSocket->peerAddress().toString()+": Network proxy authentication required (Not supported yet). Terminating");
    logEmbossRun(kembossServer::ClientServerConnectionFailed);
    closeConnection();
}

void KembossThread::tcpError(QAbstractSocket::SocketError error)
{
    if (error != QAbstractSocket::RemoteHostClosedError) {
        msgStdErr(tcpSocket->peerAddress().toString()+": TCP error: "+tcpSocket->errorString());
        logEmbossRun(kembossServer::ClientServerConnectionFailed);
        closeConnection();
    } else {
        qDebug() << "Remote host closed connection";
    }
}

void KembossThread::peerVerifyError (const QSslError &error)
{
    if (error != QSslError::NoError) {
        qDebug() << "Peer verify error";
        // With Qt5 this signal is emitted even if we set the PeerVerifyMode to none
        msgStdErr(tcpSocket->peerAddress().toString()+": peerVerifyError: "+error.errorString());
        logEmbossRun(kembossServer::ClientServerConnectionFailed);
    }
}

void KembossThread::sslErrors (const QList<QSslError> &errors)
{
    foreach (const QSslError error, errors) {
        if (error != QSslError::NoError) {
            msgStdErr(tcpSocket->peerAddress().toString()+": sslErrors: "+error.errorString());
            logEmbossRun(kembossServer::ClientServerConnectionFailed);
        }
    }
    tcpSocket->ignoreSslErrors();
}

void KembossThread::embossProcessError(QProcess::ProcessError error)
{
    qDebug() << "EMBOSS process error";
    msgStdOut(tcpSocket->peerAddress().toString()+": EMBOSS program failed.");
    switch (error) {
        case QProcess::FailedToStart: msgStdErr(tcpSocket->peerAddress().toString()+": EMBOSS program failed to start");
                                      logEmbossRun(kembossServer::FailedToStart);
                                      break;
        case QProcess::Crashed: msgStdErr(tcpSocket->peerAddress().toString()+": EMBOSS program crashed");
                                logEmbossRun(kembossServer::Crashed);
                                break;
        case QProcess::Timedout: msgStdErr(tcpSocket->peerAddress().toString()+": EMBOSS program timed out");
                                 logEmbossRun(kembossServer::Timedout);
                                 break;
        case QProcess::WriteError: msgStdErr(tcpSocket->peerAddress().toString()+": EMBOSS program write error");
                                   logEmbossRun(kembossServer::WriteError);
                                   break;
        case QProcess::ReadError: msgStdErr(tcpSocket->peerAddress().toString()+": EMBOSS program read error");
                                  logEmbossRun(kembossServer::ReadError);
                                  break;
        case QProcess::UnknownError: msgStdErr(tcpSocket->peerAddress().toString()+": Unknown error on EMBOSS program execution");
                                     logEmbossRun(kembossServer::UnknownErrorOnEmbossRun);
    }
    closeConnection();
}
