/***************************************************************************
 *                                                                         *
 *   kemboss - Graphical user interface for EMBOSS                         *
 *                                                                         *
 *   Copyright (C) 2007-2019 Tobias Schwarz (mail@tobias-schwarz.net)      *
 *                                                                         *
 *   Based on the EMBOSS.kaptn scripts by Thomas Siegmund                  *
 *                                                                         *
 *   This program is free software; you can redistribute and/or modify     *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version, and provided that the above       *
 *   copyright and permission notice is included with all distributed      *
 *   copies of this or derived software.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 **************************************************************************/

#include "sql.h"
#include "simplecrypt.h"
#include "ksshared.h"
#include "../main.h"

#include <QtDebug>
#include <QtSql/QSqlQuery>
#include <QtSql/QSqlError>
#include <QtSql/QSqlRecord>
#include <QtSql/QSqlQueryModel>
#include <QDateTime>
#include <QSettings>
#include <QStringList>
#include <QMutexLocker>

ksql::dbType ksql::checkDbDriver()
{
    qDebug() << "Checking Db driver...";

    QString configFile = kembossServer::getServerConfigFile();
    QSettings settings(configFile, QSettings::IniFormat);
    settings.beginGroup("Database");
    QString dbTypeStr = "Q" + settings.value("DatabaseType", "MYSQL").toString();
    settings.endGroup();

    if (!QSqlDatabase::isDriverAvailable(dbTypeStr)) {
        kembossServer::msgStdErr("The driver plugin " + dbTypeStr + " is not available. "
                   "The plugins have to be in the subdirectory 'sqldrivers' of the application directory.\n"
                   "Available drivers:");
        QStringList availableDrivers = QSqlDatabase::drivers();
        foreach (const QString driver, availableDrivers) {
            kembossServer::msgStdErr(driver);
        }
        return dbType(unknown);
    }

    qDebug() << dbTypeStr << "driver loaded";
    if (dbTypeStr == "QMYSQL")
        return dbType(MySQL);
    else if (dbTypeStr == "QSQLITE")
        return dbType(SQLite);
    else return dbType(unknown);
}

bool ksql::connectToDb(QSqlDatabase *db, const dbType type)
{
    qDebug() << "Connecting to SQL database...";

    QSqlError error;

    QString configFile = kembossServer::getServerConfigFile();
    QSettings settings(configFile, QSettings::IniFormat);

    if (type == dbType(MySQL)) {
        settings.beginGroup("MySQL");
        db->setHostName(settings.value("DatabaseHostname", "localhost").toString());
        db->setPort(settings.value("DatabasePort", 3306).toInt());
        db->setDatabaseName(settings.value("DatabaseName", "kemboss").toString());
        db->setUserName(settings.value("DatabaseUser", "kemboss").toString());
        SimpleCrypt crypto(Q_UINT64_C(0x1c3ad4a3acb9f023));
        db->setPassword(crypto.decryptToString(settings.value("DatabasePassword").toString()));
        settings.endGroup();
    } else if (type == dbType(SQLite)) {
        settings.beginGroup("SQLite");
        db->setDatabaseName(settings.value("DatabaseName", "kemboss").toString());
        settings.endGroup();
    }

    if (!db->open()) {
        error = db->lastError();
        kembossServer::msgStdErr("Unable to open the SQL database: " + error.text());
        return false;
    }

    return true;
}

bool ksql::createDbTables(const dbType type)
{
    QString createServerTable;
    QString createUserTable;
    QString createInputFilesTable;
    QString createProgramRunTable;
    switch (type) {
    case unknown:
        return false;
    case MySQL:
        createServerTable = "CREATE TABLE IF NOT EXISTS `server` ("
                "`version` smallint(5) unsigned NOT NULL"
                ") ENGINE=InnoDB DEFAULT CHARSET=utf8;";
        createUserTable = "CREATE TABLE IF NOT EXISTS `user` ("
                "`id` int(10) unsigned NOT NULL AUTO_INCREMENT,"
                "`name` tinytext NOT NULL,"
                "`group` tinytext NOT NULL,"
                "`pw` tinytext NOT NULL,"
                "`pw_change_required` BOOLEAN NOT NULL DEFAULT FALSE,"
                "`email` tinytext NOT NULL,"
                "`created` datetime NOT NULL,"
                "`last_login` datetime NOT NULL,"
                "PRIMARY KEY (`id`)"
                ") ENGINE=InnoDB DEFAULT CHARSET=utf8 AUTO_INCREMENT=1 ;";
        createInputFilesTable = "CREATE TABLE IF NOT EXISTS `input_files` ("
                "`id` int(10) unsigned NOT NULL AUTO_INCREMENT,"
                "`user` int(10) unsigned NOT NULL,"
                "`file` text NOT NULL,"
                "`md5` char(32) NOT NULL,"
                "`added` timestamp NOT NULL DEFAULT CURRENT_TIMESTAMP,"
                "`last_accessed` datetime NOT NULL,"
                "PRIMARY KEY (`id`)"
                ") ENGINE=InnoDB DEFAULT CHARSET=utf8 AUTO_INCREMENT=1 ;";
        createProgramRunTable = "CREATE TABLE IF NOT EXISTS `program_run` ("
                "`id` int(10) unsigned NOT NULL AUTO_INCREMENT,"
                "`user_id` int(10) unsigned NOT NULL,"
                "`ip` tinytext NOT NULL,"
                "`program` tinytext NOT NULL,"
                "`run_at` timestamp NOT NULL DEFAULT CURRENT_TIMESTAMP,"
                "`runtime` int(10) unsigned NOT NULL,"
                "`success` tinyint(1) NOT NULL,"
                "`client_version` mediumint(8) unsigned NOT NULL,"
                "`client_os` smallint(5) unsigned NOT NULL,"
                "PRIMARY KEY (`id`)"
                ") ENGINE=InnoDB DEFAULT CHARSET=utf8 AUTO_INCREMENT=1 ;";
        break;
    case SQLite:
        createServerTable = "CREATE TABLE IF NOT EXISTS `server` ("
                "`version` smallint(5) NOT NULL"
                ");";
        createUserTable = "CREATE TABLE IF NOT EXISTS `user` ("
                "`id` integer NOT NULL PRIMARY KEY AUTOINCREMENT,"
                "`name` tinytext NOT NULL,"
                "`group` tinytext NOT NULL,"
                "`pw` tinytext NOT NULL,"
                "`pw_change_required` BOOLEAN NOT NULL DEFAULT FALSE,"
                "`email` tinytext NOT NULL,"
                "`created` datetime NOT NULL,"
                "`last_login` datetime NOT NULL"
                ");";
        createInputFilesTable = "CREATE TABLE IF NOT EXISTS `input_files` ("
                "`id` integer NOT NULL PRIMARY KEY AUTOINCREMENT,"
                "`user` int(10) NOT NULL,"
                "`file` text NOT NULL,"
                "`md5` char(32) NOT NULL,"
                "`added` timestamp NOT NULL DEFAULT CURRENT_TIMESTAMP,"
                "`last_accessed` datetime NOT NULL"
                ");";
        createProgramRunTable = "CREATE TABLE IF NOT EXISTS `program_run` ("
                "`id` integer NOT NULL PRIMARY KEY AUTOINCREMENT,"
                "`user_id` int(10) NOT NULL,"
                "`ip` tinytext NOT NULL,"
                "`program` tinytext NOT NULL,"
                "`run_at` timestamp NOT NULL DEFAULT CURRENT_TIMESTAMP,"
                "`runtime` int(10) NOT NULL,"
                "`success` tinyint(1) NOT NULL,"
                "`client_version` mediumint(8) NOT NULL,"
                "`client_os` smallint(5) NOT NULL"
                ");";
        break;
    case PostgreSQL:
        break;
    case DB2:
        break;
    case IBASE:
        break;
    case OCI:
        break;
    case ODBC:
        break;
    }

    if (ksql::execSqlQuery(createServerTable,
               "Error in creating the 'server' database table.").toInt() == 0)
        return false;

    if (ksql::execSqlQuery("INSERT INTO server (version) VALUES ("+QString(KEMBOSS_SERVER_MAJOR_VERSION)+QString(KEMBOSS_SERVER_MINOR_VERSION)+QString(KEMBOSS_SERVER_MINIMUM_VERSION)+")",
                     "Error in writing the server version to the database.").toInt() == 0)
        return false;

    if (ksql::execSqlQuery(createUserTable,
               "Error in creating the 'user' database table.").toInt() == 0)
        return false;

    if (ksql::execSqlQuery(createInputFilesTable,
               "Error in creating the 'input_files' database table.").toInt() == 0)
        return false;

    if (ksql::execSqlQuery(createProgramRunTable,
               "Error in creating the 'program_run' database table.").toInt() == 0)
        return false;

    return true;
}

bool ksql::updateDbTables(const uint dbVersion)
{
    if (dbVersion < 110) {
        kembossServer::msgStdOut("Updating the database to server version 1.1.0");
        if (ksql::execSqlQuery("ALTER TABLE `program_run` ADD `client_os` SMALLINT UNSIGNED NOT NULL;",
                               "An error occurred in updating the database").toInt() == 0)
            return false;
    }

    if (ksql::execSqlQuery("UPDATE server SET version="+QString(KEMBOSS_SERVER_MAJOR_VERSION)+QString(KEMBOSS_SERVER_MINOR_VERSION)+QString(KEMBOSS_SERVER_MINIMUM_VERSION),
                           "An error occurred in updating the database").toInt() == 0)
        return false;

    return true ;
}

bool ksql::checkDb()
{
    qDebug() << "Checking SQL database...";

    bool next = true;
    dbType type;
    uint dbVersion = 0;
    bool createTables = false;
    bool updateTables = false;

    {
        QSqlDatabase db;

        type = checkDbDriver();

        db = QSqlDatabase::addDatabase(dbTypeStr(type), "kembossd");

        if (!connectToDb(&db, type)) {
            kembossServer::msgStdErr("Error in connecting to database: " + db.lastError().text());
            next = false;
        }

        if (next)
            if (!db.isValid()) {
                kembossServer::msgStdErr("Error in adding the database to the list of database connections: " + db.lastError().text());
                next = false;
            }

        if (next) {
            if (db.tables().count() == 0) {
            // If the database is empty, create the required database tables
                createTables = true;
            } else {
            // Check if version of SQL database is equivalent to the current server version, otherwise update database
                qDebug() << "Checking database version";
                QSqlQuery query(db);
                if (!query.exec("SELECT version FROM server")) {
                    kembossServer::msgStdErr(query.lastError().text());
                    kembossServer::msgStdErr(db.lastError().text());
                    kembossServer::msgStdErr("Error in retrieving the database version");
                    next = false;
                }
                if (next && !query.isActive()) {
                    kembossServer::msgStdErr("The SQL statement has not been executed successfully.");
                    next = false;
                }
                if (next && !query.first()) {
                    kembossServer::msgStdErr("The SQL statement did not return a value for the server version.");
                    next = false;
                }
                if (next) {
                    dbVersion = query.value(0).toUInt();

                    if (dbVersion == 0) {
                        kembossServer::msgStdErr("Could not retrieve the version of the database tables");
                        next = false;
                    } else if (dbVersion < QByteArray(QByteArray(KEMBOSS_SERVER_MAJOR_VERSION)+QByteArray(KEMBOSS_SERVER_MINOR_VERSION)+QByteArray(KEMBOSS_SERVER_MINIMUM_VERSION)).toUInt()) {
                    // Database tables need an update
                        updateTables = true;
                    }
                }
            }
        }
        db.close();
    }

    QSqlDatabase::removeDatabase("kembossd");

    if (next && createTables) {
        kembossServer::msgStdOut("Database is empty. Creating database tables.");
        if (!ksql::createDbTables(type))
            next = false;
    }
    if (next && updateTables)
        updateDbTables(dbVersion);

    if (!next)
        return false;

    kembossServer::msgStdOut("Connection to SQL database successful");

    return true;
}

QVariant ksql::execSqlQuery(const QString sqlQuery, const QString errorMsg, const bool returnResult)
{
    qDebug() << "Executing SQL command: " << sqlQuery;

    QMutex mutex;
    QMutexLocker locker(&mutex);

    QVariant result;

    bool next = true;

    {
        QSqlDatabase db;

        dbType type = checkDbDriver();

        db = QSqlDatabase::addDatabase(dbTypeStr(type), "kembossd");
        if (!connectToDb(&db, type)) {
            next = false;
        }

        if (next)
            if (!db.isValid()) {
                kembossServer::msgStdErr("Error in adding the database to the list of database connections: " + db.lastError().text());
                next = false;
            }
        if (next) {
            QSqlQuery query(db);
            if (!query.prepare(sqlQuery)) {
                kembossServer::msgStdErr(query.lastError().text());
                kembossServer::msgStdErr(db.lastError().text());
                kembossServer::msgStdErr(errorMsg);
                next = false;
            }
            if (!query.exec()) {
                kembossServer::msgStdErr(query.lastError().text());
                kembossServer::msgStdErr(db.lastError().text());
                kembossServer::msgStdErr(errorMsg);
                next = false;
            }
            if (next) {
                if (returnResult) {
                    if (!query.isActive()) {
                        kembossServer::msgStdErr("The SQL statement has not been executed successfully.");
                        next = false;
                    }
                    if (next && !query.isSelect()) {
                        kembossServer::msgStdErr("The SQL statement is no SELECT statement, so no result can be retrieved.");
                        next = false;
                    }
                    if (next && !query.first()) {
                        qDebug() << "The SQL query did not return a result";
                        next = false;
                        // If the SQL query does not return a result, no error message is returned, as an empty result might be correct (e.g. the user does not exist, etc.)
                    }
                    if (next) {
                        result = query.value(0);
                    }
                } else
                    result = 1; // A return value of 1 indicates success
            }
        }
        db.close();
    }

    QSqlDatabase::removeDatabase("kembossd");

    if (!next)
        result = 0;
    return(result);
}

quint32 ksql::newUser(const QString name, const QString group, const QByteArray pwHash, const bool newPw, const QString email)
{
    QMutex mutex;
    QMutexLocker locker(&mutex);

    QString pwChange;
    if (newPw) pwChange = "1"; else pwChange = "0";

    QString timestampStr = QDateTime::currentDateTime().toString("yyyy-MM-dd hh:mm:ss");

    QString SQLQuery = "INSERT INTO user (`name`, `group`, `pw`, `pw_change_required`, `email`, `created`, `last_login`) "
            "VALUES ('"+name+"', '"+group+"', '"+pwHash+"', "+pwChange+", '"+email+"', '"+timestampStr+"', '"+timestampStr+"')";

    if (ksql::execSqlQuery(SQLQuery, "Error in writing the user data to the database.").toInt() == 0) {
        kembossServer::msgStdErr("Error in adding new user to the database.", "stderr");
        return 0;
    }
    quint32 newUserId = ksql::execSqlQuery("SELECT id FROM user WHERE name = '"+name+"' AND pw = '"+pwHash+"'",
                                       "Error in getting the id of the added user.", true).toUInt();

    qDebug() << "Created new user " << newUserId;

    return newUserId;
}

QString ksql::dbTypeStr(const dbType type)
{
    switch (type) {
    case 0:
        return "unknown";
    case 1:
        return "QMYSQL";
    case 2:
        return "QSQLITE";
    case 3:
        return "QPSQL";
    case 4:
        return "QODBC";
    case 5:
        return "QDB2";
    case 6:
        return "QIBASE";
    case 7:
        return "QOCI";
    }
    return "";
}
