/***************************************************************************
 *                                                                         *
 *   kemboss - Graphical user interface for EMBOSS                         *
 *                                                                         *
 *   Copyright (C) 2007-2014 Tobias Schwarz (mail@tobias-schwarz.net)      *
 *                                                                         *
 *   Based on the EMBOSS.kaptn scripts by Thomas Siegmund                  *
 *                                                                         *
 *   This program is free software; you can redistribute and/or modify     *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version, and provided that the above       *
 *   copyright and permission notice is included with all distributed      *
 *   copies of this or derived software.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 **************************************************************************/

#include <QMessageBox>
#include <QNetworkConfiguration>
#include <QNetworkConfigurationManager>
#include <QNetworkSession>
#include <QtGlobal>
#include "main.h"
#include "shared.h"
#include "runEmbossAppOnServer.h"
#include "programWindowClass.h"
#include "changePassword.h"
#include "certificateinfo.h"
#include "ui_sslerrors.h"

krunEmbossAppOnServer::krunEmbossAppOnServer(QWidget *parent, kembossSettings *fksettings)
    : krunEmbossApp(parent, fksettings) //, networkSession(nullptr)
{

}

bool krunEmbossAppOnServer::runApp()
{
   krunEmbossApp::runApp();
   // Do the basic preparations for running the EMBOSS app as defined in the krunEmbossApp base class

   fileListID = 0;
   blockSize = 0;
   if (!startKembossClient()) return false;
   return true;
}

bool krunEmbossAppOnServer::startKembossClient()
{
    tcpSocket = new QSslSocket(this);

    // Check if SSL is supported by the installed Qt libraries
    if (!tcpSocket->supportsSsl()) {
        programProgress->ProgramOutputTextEdit->append("System does not support SSL. Terminating...");
        return false;
    }

    connect(tcpSocket, SIGNAL(readyRead()), this, SLOT(readData()));
    connect(tcpSocket, SIGNAL(proxyAuthenticationRequired(QNetworkProxy, QAuthenticator*)), this, SLOT(proxyAuthenticationRequired(QNetworkProxy, QAuthenticator*)));
    connect(tcpSocket, SIGNAL(error(QAbstractSocket::SocketError)), this, SLOT(displayError(QAbstractSocket::SocketError)));
    connect(tcpSocket, SIGNAL(peerVerifyError(const QSslError &error)), this, SLOT(checkPeerVerifyError(const QSslError &error)));
    connect(tcpSocket, SIGNAL(sslErrors(const QList<QSslError>)), this, SLOT(checkSslErrors(const QList<QSslError>)));
    connect(tcpSocket, SIGNAL(hostFound()), this, SLOT(kembossHostFound()));
    connect(tcpSocket, SIGNAL(connected()), this, SLOT(connectionEstablished()));
    connect(tcpSocket, SIGNAL(encrypted()), this, SLOT(socketEncrypted()));
    connect(tcpSocket, SIGNAL(disconnected()), this, SLOT(connectionClosed()));

    QNetworkConfigurationManager manager;
    if (manager.capabilities() & QNetworkConfigurationManager::NetworkSessionRequired) {
        programProgress->ProgramOutputTextEdit->append("Opening network session...");
        // Get saved network configuration
        QSettings *settings;
        if (ksettings->portableApp()) {
            settings = new QSettings(ksettings->kembossPortableConfigDir()+"NetworkConfiguration", QSettings::IniFormat);
        } else {
            settings = new QSettings(QSettings::UserScope, "kemboss.org", "NetworkConfiguration");
        }
        settings->beginGroup(QLatin1String("QtNetwork"));
        const QString id = settings->value(QLatin1String("DefaultNetworkConfiguration")).toString();
        settings->endGroup();

        // If the saved network configuration is not currently discovered use the system default
        QNetworkConfiguration config = manager.configurationFromIdentifier(id);
        if ((config.state() & QNetworkConfiguration::Discovered) !=
            QNetworkConfiguration::Discovered) {
            config = manager.defaultConfiguration();
        }

        networkSession = new QNetworkSession(config, this);
        connect(networkSession, SIGNAL(opened()), this, SLOT(sessionOpened()));
        networkSession->open();
    }

    programProgress->ProgramOutputTextEdit->append("\nUsing SSL library " + tcpSocket->sslLibraryVersionString());

    QString PortStr;
    PortStr.setNum(ksettings->port());
    programProgress->ProgramOutputTextEdit->append("Trying to connect to "+ksettings->hostname()+" on port "+PortStr);

    // Start SSL encrypted connection with server
    tcpSocket->connectToHostEncrypted(ksettings->hostname(), ksettings->port(), QIODevice::ReadWrite);

    return true;
}

void krunEmbossAppOnServer::sessionOpened()
{
    qDebug() << "Network session opened";
    // Save the used configuration
    QNetworkConfiguration config = networkSession->configuration();
    QString id;
    if (config.type() == QNetworkConfiguration::UserChoice)
        id = networkSession->sessionProperty(QLatin1String("UserChoiceConfiguration")).toString();
    else
        id = config.identifier();
    QSettings *settings;
    if (ksettings->portableApp()) {
        settings = new QSettings(ksettings->kembossPortableConfigDir()+"NetworkConfiguration", QSettings::IniFormat);
    } else {
        settings = new QSettings(QSettings::UserScope, "kemboss.org", "NetworkConfiguration");
    }
    settings->beginGroup(QLatin1String("QtNetwork"));
    settings->setValue(QLatin1String("DefaultNetworkConfiguration"), id);
    settings->endGroup();

    programProgress->ProgramOutputTextEdit->append("Network session opened.");
}

void krunEmbossAppOnServer::kembossHostFound()
{
    qDebug() << "Kembossd host found";
    programProgress->ProgramOutputTextEdit->append("Host found.");
}

void krunEmbossAppOnServer::connectionEstablished()
{
    qDebug() << "Connection to server established";
    programProgress->ProgramOutputTextEdit->append("Connection to server established.");
}

void krunEmbossAppOnServer::socketEncrypted()
{
    qDebug() << "Trying to establish encrypted connection";
    QSslCipher ciph = tcpSocket->sessionCipher();
    if (ciph.isNull()) {
        qDebug() << "Connection is not encrypted! Terminating.";
        programProgress->ProgramOutputTextEdit->append("Connection is not encrypted! Terminating.");
        killProcess();
    }
    QString cipher = QString("%1, %2 (%3/%4)").arg(ciph.authenticationMethod())
                     .arg(ciph.name()).arg(ciph.usedBits()).arg(ciph.supportedBits());
    qDebug() << "SSL connection established. Cryptographic cipher: " << cipher;
    programProgress->ProgramOutputTextEdit->append("SSL connection established.\n"
                                                   "Cryptographic cipher: "+cipher);
}

void krunEmbossAppOnServer::readData()
{
    // First, read all incoming data from the server.
    QDataStream in(tcpSocket);
    in.setVersion(QDataStream::Qt_4_6);

    // The servers response starts with the size of the packet, so first we need to ensure that we can read the size
    if (blockSize == 0) {
        if (tcpSocket->bytesAvailable() < (int)sizeof(qint64))
            return;
        in >> blockSize;
    }

    // Then we will wait until QTcpSocket has received the complete command of the server
    if (tcpSocket->bytesAvailable() < blockSize)
        return;

    // If we have received the complete command of the server, it is read into a QByteArray
    QByteArray incoming;
    in >> incoming;
    qDebug() << "Receiving: " << incoming;

    // The first 5 characters of the incoming data are the command.
    QByteArray command = incoming.left(5);
    incoming.remove(0, 5);

    if (command == "[sst]") { // Receive the server status
        readServerStatus(QString::fromUtf8(incoming));
    } else if (command == "[scf]") { // Establishing an SSL connection was not possible. We have to ask the user what to do

    } else if (command == "[uak]") { // User acknowledged
        programProgress->ProgramOutputTextEdit->append("User information verified");
        writeEmbossConfiguration();
    } else if (command == "[ude]") { // User denied
        programProgress->ProgramOutputTextEdit->append("Access to the server denied. Please check Your user account");
        processFinished(-1);
    } else if (command == "[uwp]") { // Username was found in the database but the entered password was wrong
        programProgress->ProgramOutputTextEdit->append("The username exists but the entered password was wrong. You either mistyped Your password or You have to choose a different username\n");
        processFinished(-1);
    } else if (command == "[npw]") { // User acknowledged, password expired, update password
        updatePw();
    } else if (command == "[iep]") { // Receiving installed EMBASSY packages
        readInstalledEmbassyPackages(incoming);
    } else if (command == "[eea]") { // Execution of EMBOSS binary on server allowed. Send the EMBOSS output file list
        writeEmbossOutputFileList();
    } else if (command == "[eor]") { // EMBOSS output file list received. The client has to send an EMBOSS input file ID or [rap] if all input files are sent
        writeData(QString(checkAllFilesTransferred()).toUtf8());
    } else if (command == "[has]") { // EMBOSS input file is already stored on the server, no need to send it
        programProgress->ProgramOutputTextEdit->append("EMBOSS input file is already cached on the server");
        // either the next input file ID or [rap] is send to the server
        fileListID++;
        writeData(QString(checkAllFilesTransferred()).toUtf8());
    } else if (command == "[get]") { // Send EMBOSS input file to server
        writeEmbossInputFile(cembossInputFiles.at(fileListID));
    } else if (command == "[fin]") { // EMBOSS input file transfer succeeded, send next input file ID or [rap]
        programProgress->ProgramOutputTextEdit->append("EMBOSS input file transfer succeeded");
        // either the next input file ID or [rap] is send to the server
        fileListID++;
        writeData(QString(checkAllFilesTransferred()).toUtf8());
    } else if (command == "[pna]") { // execution of program not allowed
        programProgress->ProgramOutputTextEdit->append("Execution of program on the server not allowed");
        processFinished(-1);
    } else if (command == "[bnf]") { // EMBOSS binary not found on the server
        programProgress->ProgramOutputTextEdit->append("EMBOSS binary not found on the server");
        processFinished(-1);
    } else if (command == "[eou]") { // EMBOSS output on stdout
        readEmbossOutput(incoming);
    } else if (command == "[fod]") { // EMBOSS output file identifier (filename (without path) followed by the md5 sum of the file)
        readEmbossOutputFileID(incoming);
    } else if (command == "[eof]") { // gzipped EMBOSS output file
        readEmbossOutputFile(incoming);
    } else if (command == "[tof]") { // Transfer of EMBOSS output files finished / program execution finished
        processFinished(0);
    } else {
        programProgress->ProgramOutputTextEdit->append("Received unknown command");
        processFinished(-1);
    }

    // The blockSize is reset to 0 so that the next command can be retrieved from the server
    blockSize = 0;
}

qint64 krunEmbossAppOnServer::writeData(QByteArray data)
{
    qDebug() << "Writing: " << data;
    QByteArray block;
    QDataStream out(&block, QIODevice::WriteOnly);
    out.setVersion(QDataStream::Qt_4_6);
    out << (qint64)0;
    out << data;
    out.device()->seek(0);
    out << (qint64)(block.size() - sizeof(qint64));
    return tcpSocket->write(block);
}

void krunEmbossAppOnServer::readServerStatus(const QString &data)
{
    QStringList datalist = data.split("[spl]");
    int serverStatus;
    serverStatus = datalist[0].toInt();
    programProgress->ProgramOutputTextEdit->append("\nServer status is: "+kemboss::serverStatusString(serverStatus));
    programProgress->ProgramOutputTextEdit->append(datalist[1]+"\n"); // Write server welcome message to the program progress window
    // If the server status is 'Ready', the client sends the user data
    // Otherwise the connection is terminated.
    if (serverStatus == kemboss::Ready ) {
        writeUser();
    } else {
        processFinished(-1);
    }
}

void krunEmbossAppOnServer::writeUser()
{
    programProgress->ProgramOutputTextEdit->append("Sending kemboss version number and user information.");
    QByteArray command = "[cvn]"+QByteArray(KEMBOSS_MAJOR_VERSION)+QByteArray(KEMBOSS_MINOR_VERSION)+QByteArray(KEMBOSS_MINIMUM_VERSION)
       +"[spl]"+ksettings->userName().toUtf8()
       +"[spl]"+ksettings->userGroup().toUtf8()
       +"[spl]"+ksettings->userEmail().toUtf8()
       +"[spl]"+ksettings->userPwHash()
       +"[spl]0"
       #ifdef Q_OS_BSD4
           "1"
       #endif
       #ifdef Q_OS_LINUX
           "2"
       #endif
       #ifdef Q_OS_MAC
           "3"
       #endif
       #ifdef Q_OS_Solaris
           "4"
       #endif
       #ifdef Q_OS_WIN32
           "5"
       #endif
       #ifdef Q_OS_UNIX
           "6"
       #endif
       ;
    writeData(command);
}

void krunEmbossAppOnServer::writeEmbossConfiguration()
{
    programProgress->ProgramOutputTextEdit->append("Sending EMBOSS program name and configuration");
    QString command = "[epn]"+embossApp()+"[spl]"+programConfiguration.join("[spl]");
    writeData(QString(command).toUtf8());
}

void krunEmbossAppOnServer::writeEmbossOutputFileList()
{
    programProgress->ProgramOutputTextEdit->append("Sending EMBOSS output file list");
    QString command = "[eol]" + cembossOutputFiles.join("[spl]");
    writeData(QString(command).toUtf8());
}

QString krunEmbossAppOnServer::checkAllFilesTransferred()
{
    if (fileListID < cembossInputFiles.size()) {
        if (QFile::exists(cembossInputFiles.at(fileListID))) {
            return(embossInputFileID(cembossInputFiles.at(fileListID)));
        } else {
            fileListID++;
            return(checkAllFilesTransferred());
        }
    } else {
        // When all input files are transferred to the server, the client asks the server to run the EMBOSS program
        fileListID = 0;
        return("[rap]");
    }
}

QString krunEmbossAppOnServer::embossInputFileID(const QString &filename)
{
    programProgress->ProgramOutputTextEdit->append("Sending ID of EMBOSS input file " + filename);

    return("[fid]" + filename + "[spl]" + QString::fromUtf8(kemboss::getFileMD5Sum(filename)));
}

void krunEmbossAppOnServer::writeEmbossInputFile(const QString &filename)
{
    programProgress->ProgramOutputTextEdit->append("Sending EMBOSS input file "+filename);

    QFile file(filename);
    file.open(QFile::ReadOnly);
    writeData("[eif]" + qCompress(file.readAll()));
}

void krunEmbossAppOnServer::readEmbossOutput(const QByteArray &data)
{
    cprogramOutput = data;
    krunEmbossApp::updateProgramProgressForm();
    writeData("[ore]");
}

void krunEmbossAppOnServer::readEmbossOutputFileID(const QByteArray &data)
{
    //programProgress->ProgramOutputTextEdit->append("Receiving EMBOSS output file identifier");

    int pos = data.indexOf("[spl]",0);

    // The filename (without path) is received
    outputFilename = QString(data.left(pos)).toUtf8();

    // The md5sum of the output file is received
    outputMD5sum = data.mid(pos+5, -1);
    writeData("[oir]");
}

void krunEmbossAppOnServer::readEmbossOutputFile(const QByteArray &data)
{
    programProgress->ProgramOutputTextEdit->append("Receiving EMBOSS output file " + outputFilename);

    // Search for the correct directory to save the output file in
    // EMBOSS often appends something to the output files so we extract the basename of the output file
    QString basename;
    QFileInfo fileInfo(outputFilename);
    basename = fileInfo.baseName();

    // We then search for the basename in the embossOutputFilesList and extract the path
    QStringList result;
    result = cembossOutputFiles.filter(basename, Qt::CaseInsensitive);
    // If we do not find the basename in the embossOutputFilesList we take the first entry of the list, as it contains the path of the OutputFilesLineEdit
    if (result.isEmpty()) {
        fileInfo.setFile(cembossOutputFiles.at(0));
    } else {
        fileInfo.setFile(result.at(0));
    }
    QString filename = fileInfo.absolutePath()+"/"+outputFilename;

    // Then we have to check if a file with the outputFilename already exists
    quint16 fileNum = 0;
    QString origFilename = filename;
    while (QFile::exists(filename)) {
        // If a file with the filename already exists, a number is appended to the filename until the file doesn't exist
        fileNum++;
        QString fileNumStr;
        fileNumStr.setNum(fileNum);
        filename = origFilename + "." + fileNumStr;
    }

    // The output file is received, uncompressed and saved
    QFile file(filename);
    if (!file.open(QIODevice::WriteOnly)) {
        programProgress->ProgramOutputTextEdit->append("EMBOSS output file could not be saved!");
        return;
    }
    file.write(qUncompress(data));
    file.close();

    // Following file transfer the MD5sum of the transferred file is compared to the MD5sum received from the server
    if (kemboss::getFileMD5Sum(filename) != outputMD5sum ) {
        programProgress->ProgramOutputTextEdit->append("MD5sum of the received EMBOSS output file " + outputFilename + " is not identical with the MD5sum calculated on the server."
                                                       "An error occured during data transfer!");
    }

    outputFilename = "";

    // The server is asked to send the next EMBOSS output file identifier
    writeData("[ore]");
}

void krunEmbossAppOnServer::updatePw()
{
    programProgress->ProgramOutputTextEdit->append("User information verified. Your password has expired. Please enter a new password");
    ChangePasswordDlg ChangePwDialog(ksettings);
    ChangePwDialog.exec();
    writeData("[pwu]"+ksettings->userPwHash());
}

void krunEmbossAppOnServer::readInstalledEmbassyPackages(const QByteArray &data)
{
    bool installed;
    if (data == "p1") installed = true; else installed = false;
    QSettings *settings;
    if (ksettings->portableApp()) {
        settings = new QSettings(ksettings->kembossPortableConfigDir()+"kemboss", QSettings::IniFormat);
    } else {
        settings = new QSettings(QSettings::UserScope, "kemboss.org", "kemboss");
    }
    settings->setValue("phylipnew_installed", installed);
    if (installed) {
        programProgress->ProgramOutputTextEdit->append("kemboss server supports PHYLIPNEW");
    } else {
        programProgress->ProgramOutputTextEdit->append("kemboss server does not support PHYLIPNEW");
    }
    processFinished(0);
}

void krunEmbossAppOnServer::processFinished(int exitCode)
{
    tcpSocket->disconnectFromHost();
    krunEmbossApp::processFinished(exitCode);
}

void krunEmbossAppOnServer::killProcess()
{
    writeData("[cpe]");
    tcpSocket->abort();
    krunEmbossApp::killProcess();
}

void krunEmbossAppOnServer::connectionClosed()
{
    programProgress->ProgramOutputTextEdit->append("Connection closed");
}

void krunEmbossAppOnServer::displayError(QAbstractSocket::SocketError error) {
    if (error == QAbstractSocket::RemoteHostClosedError)
        return;

    switch (error) {
    case QAbstractSocket::HostNotFoundError:
        QMessageBox::information(parentWindow, tr("TCP socket error"),
                                 tr("The host was not found. Please check the "
                                    "host name and port settings."));
        break;
    default:
        QMessageBox::information(parentWindow, tr("TCP socket error"), tcpSocket->errorString());
    }
    processFinished(-1);
}

void krunEmbossAppOnServer::checkPeerVerifyError(const QSslError &error)
{
    qDebug() << "SSL error occured.";
    if (error != QSslError::NoError) {
        QMessageBox::information(parentWindow, tr("SSL error"),
                                 tr("SSL error occured: ")+error.errorString());
    }
}

void krunEmbossAppOnServer::checkSslErrors(const QList<QSslError> &errors)
{
    qDebug() << "SSL error occured.";
    // The SslError signal sometimes is invoked only due to a "NoError". We do not want to bother the user because of this
    bool realError = false;
    foreach (const QSslError &error, errors) {
        if (error != QSslError::NoError)
            realError = true;
    }
    if (!realError) {
        tcpSocket->ignoreSslErrors();
        return;
    }

    // Check if the host is already saved in the list of hosts whose SSL errors are always ignored
    QSettings *settings;
    if (ksettings->portableApp()) {
        settings = new QSettings(ksettings->kembossPortableConfigDir()+"kemboss", QSettings::IniFormat);
    } else {
        settings = new QSettings(QSettings::UserScope, "kemboss.org", "kemboss");
    }
    settings->beginGroup("AllowedSSLErrorsHosts");
    int size = settings->value("size", 0).toInt();
    QString iString;
    for (int i = 1; i <= size; ++i) {
        iString.setNum(i);
        if (settings->value("Hostname"+iString).toString() == ksettings->hostname()) {
            qDebug() << "SSL error is ignored for this host";
            tcpSocket->ignoreSslErrors();
            settings->endGroup();
            return;
        }
    }
    settings->endGroup();

    // Otherwise we have to ask the user what to do
    QDialog errorDialog(programProgress);
    Ui_SslErrors ui;
    ui.setupUi(&errorDialog);
    connect(ui.certificateChainButton, SIGNAL(clicked()), this, SLOT(displayCertificateInfo()));
    connect(ui.ignoreHostButton, SIGNAL(clicked()), this, SLOT(ignoreHostSSLErrors()));

    foreach (const QSslError &error, errors) {
        if (error != QSslError::NoError)
            ui.sslErrorList->addItem(error.errorString());
    }

    if (errorDialog.exec() == QDialog::Accepted) {
        ignoreHostSSLErrors();
        tcpSocket->ignoreSslErrors();
    } else {
        processFinished(-1);
    }
}

void krunEmbossAppOnServer::ignoreHostSSLErrors()
{
    QSettings *settings;
    if (ksettings->portableApp()) {
        settings = new QSettings(ksettings->kembossPortableConfigDir()+"kemboss", QSettings::IniFormat);
    } else {
        settings = new QSettings(QSettings::UserScope, "kemboss.org", "kemboss");
    }
    settings->beginGroup("AllowedSSLErrorsHosts");
    int size = settings->value("size", 0).toInt()+1;
    QString sizeString;
    sizeString.setNum(size);
    settings->setValue("Hostname"+sizeString, ksettings->hostname());
    settings->setValue("size", size);
    settings->endGroup();
}

void krunEmbossAppOnServer::proxyAuthenticationRequired(const QNetworkProxy &, QAuthenticator *)
{
    programProgress->ProgramOutputTextEdit->append(tr("Proxy authentication required (not supported yet). Terminating"));
    processFinished(-1);
}

void krunEmbossAppOnServer::displayCertificateInfo()
{
    CertificateInfo *info = new CertificateInfo(programProgress);
    info->setCertificateChain(tcpSocket->peerCertificateChain());
    info->exec();
    info->deleteLater();
}
